from __future__ import annotations

from typing import Any, Callable, Dict, List, Union

from pydantic import BaseModel

from .. import util
from ..fx.Graph import Graph
from ..fx.Node import Node


class NodeModel(BaseModel):
    class Reference(BaseModel):
        name: str

    name: str
    target: Union[Callable, str]
    args: List[Any]
    kwargs: Dict[str, Any]

    @staticmethod
    def from_node(node: Node):
        def _reference(node: Node):
            return NodeModel.Reference(name=node.name)

        args = util.apply(node.args, _reference, Node)
        kwargs = util.apply(node.kwargs, _reference, Node)

        return NodeModel(name=node.name, target=node.target, args=args, kwargs=kwargs)

    @staticmethod
    def to_node(graph: Graph, nodes: Dict[str, NodeModel], node_model: NodeModel):
        def _dereference(reference: NodeModel.Reference):
            return NodeModel.to_node(graph, nodes, nodes[reference.name])

        # Arguments might be interventions themselves so recurse.
        args = util.apply(node_model.args, _dereference, NodeModel.Reference)
        kwargs = util.apply(node_model.kwargs, _dereference, NodeModel.Reference)

        # Processing of args may have already created an Intervention for this node so just return it.
        if node_model.name in graph.nodes:
            return graph.nodes[node_model.name]

        graph.add(
            graph=graph,
            value=None,
            target=node_model.target,
            args=args,
            kwargs=kwargs,
            name=node_model.name,
        )



from __future__ import annotations

from typing import Callable, List, Union

import torch
from torch.utils.hooks import RemovableHandle
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoTokenizer,
    BatchEncoding,
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)

from .AbstractModel import AbstractModel


class LanguageModel(AbstractModel):
    def __init__(self, *args, **kwargs) -> None:
        self.config: PretrainedConfig = None
        self.tokenizer: PreTrainedTokenizer = None
        self.meta_model: PreTrainedModel = None
        self.local_model: PreTrainedModel = None

        super().__init__(*args, **kwargs)

    def _register_increment_hook(self, hook: Callable) -> RemovableHandle:
        return self.local_model.register_forward_hook(hook)

    def _load_meta(self, repoid_or_path, *args, **kwargs) -> PreTrainedModel:
        self.config = AutoConfig.from_pretrained(repoid_or_path, *args, **kwargs)

        self.tokenizer = AutoTokenizer.from_pretrained(
            repoid_or_path, config=self.config, padding_side="left"
        )
        self.tokenizer.pad_token = self.tokenizer.eos_token

        return AutoModelForCausalLM.from_config(self.config, trust_remote_code=True)

    def _load_local(self, repoid_or_path, *args, **kwargs) -> PreTrainedModel:
        return AutoModelForCausalLM.from_pretrained(
            repoid_or_path, *args, config=self.config, **kwargs
        )

    def _prepare_inputs(
        self,
        inputs: Union[
            str, List[str], List[List[str]], List[int], List[List[int]], torch.Tensor
        ],
    ) -> BatchEncoding:
        if isinstance(inputs, str) or (
            isinstance(inputs, list) and isinstance(inputs[0], int)
        ):
            inputs = [inputs]

        if isinstance(inputs, torch.Tensor) and inputs.ndim == 1:
            inputs = inputs.unsqueeze(0)

        if not isinstance(inputs[0], str):
            inputs = [self.tokenizer.decode(ids) for ids in inputs]

        return self.tokenizer(inputs, return_tensors="pt", padding=True)

    def _run_meta(self, inputs, *args, **kwargs) -> None:
        inputs = self._prepare_inputs(inputs)

        self.meta_model(*args, **inputs.copy().to("meta"), **kwargs)

        return inputs["input_ids"]

    def _run_local(self, inputs, *args, **kwargs):
        inputs = self._prepare_inputs(inputs)

        return self.local_model(*args, **inputs.to(self.local_model.device), **kwargs)

    def _generation(self, inputs, *args, **kwargs) -> None:
        inputs = self._prepare_inputs(inputs)

        return self.local_model.generate(
            *args, **inputs.to(self.local_model.device), **kwargs
        )



import time
from functools import wraps
from typing import Any, Callable, Type, Union

import torch

Primative = Union[str, int, float, bool]
Value = Union[Primative, torch.Tensor]


def apply(data: Any, fn: Callable, cls: type):
    if isinstance(data, cls):
        return fn(data)

    if isinstance(data, list):
        return [apply(_data, fn, cls) for _data in data]

    if isinstance(data, tuple):
        return tuple([apply(_data, fn, cls) for _data in data])

    if isinstance(data, dict):
        return {key: apply(value, fn, cls) for key, value in data.items()}

    return data


def fetch_attr(object: object, target: str):
    target_atoms = target.split(".")
    for i, atom in enumerate(target_atoms):
        object = getattr(object, atom)
    return object


def wrap(object: object, wrapper: Type, *args, **kwargs):
    if isinstance(object, wrapper):
        return object

    object.__class__ = type(object.__class__.__name__, (wrapper, object.__class__), {})

    wrapper.__init__(object, *args, **kwargs)

    return object


def timed(func, lggr):
    """This decorator prints the execution time for the decorated function."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        lggr.debug(f"Method `{func.__qualname__}` ran in {round(end - start, 6)}s")
        return result

    return wrapper


def cross_entropy_loss(
    logits: torch.Tensor,
    target_ids: torch.Tensor,
    shift: bool = False,
    avg_batch: bool = True,
    avg_token: bool = True,
):
    logits = logits.cpu()
    target_ids = target_ids.cpu()

    if logits.ndim == 2:
        logits = logits.unsqueeze(0)

    if target_ids.ndim == 1:
        target_ids = target_ids.unsqueeze(0)

    assert logits.ndim == 3
    assert target_ids.ndim == 2
    assert logits.size(0) == target_ids.size(0)
    assert logits.size(1) == target_ids.size(1)

    if shift:
        logits = logits[:, :-1]
        target_ids = target_ids[:, 1:]

    target_ids = target_ids.long()

    batch_losses = []

    for batch_idx in range(len(logits)):
        batch_loss = torch.nn.functional.cross_entropy(
            logits[batch_idx],
            target_ids[batch_idx],
            reduction="mean" if avg_token else "none",
        )
        batch_losses.append(batch_loss)

    batch_losses = torch.stack(batch_losses)

    if avg_batch:
        batch_losses = batch_losses.mean(dim=0)

    return batch_losses


class WrapperModule(torch.nn.Module):
    def forward(self, *args, **kwargs):
        if len(args) == 1:
            args = args[0]

        return args



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

from ..fx.Proxy import Proxy

if TYPE_CHECKING:
    from .Generator import Generator


class Invoker:
    def __init__(self, generator: "Generator", input, *args, **kwargs) -> None:
        self.generator = generator
        self.input = input
        self.args = args
        self.kwargs = kwargs
        self.tokens = None
        self.ids = None

    def __enter__(self) -> Invoker:
        # Were in a new invocation so set generation_idx to 0,
        self.generator.generation_idx = 0

        # Run graph_mode with meta tensors to collect shape information,
        token_ids = self.generator.model._run_meta(self.input, *self.args, **self.kwargs)

        # Decode tokenized inputs for user usage.
        self.tokens = [
            [self.generator.model.tokenizer.decode(token) for token in ids]
            for ids in token_ids
        ]
        self.ids = token_ids

        self.generator.batch_size = len(self.ids)

        # Rebuild prompt from tokens (do this becuase if they input ids directly, we still need to pass
        # all input data at once to a tokenizer to correctly batch the attention).
        self.generator.prompts.extend(["".join(tokens) for tokens in self.tokens])

        if len(self.tokens) == 1:
            self.tokens = self.tokens[0]
            self.ids = self.ids[0]

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        pass

    def next(self) -> None:
        # .next() increases which generation idx the interventions happen.
        self.generator.generation_idx += 1

        # Run graph with singe token input.
        self.generator.model._run_meta("_", *self.args, **self.kwargs)

    def save_all(self) -> Dict[str, Proxy]:
        """Saves the output of all modules and returns a dictionary of [module_path -> save proxy]

        Returns:
            Dict[str, Proxy]: _description_
        """
        result = {}

        for name, module in self.generator.model.meta_model.named_modules():
            result[module.module_path] = module.output.save()

        return result



from __future__ import annotations

from typing import Any, List, Union

import diffusers
import torch
from diffusers import AutoencoderKL, SchedulerMixin, UNet2DConditionModel
from PIL import Image
from torch.utils.hooks import RemovableHandle
from transformers import CLIPTextModel, CLIPTokenizer

from .AbstractModel import AbstractModel


class Diffuser(torch.nn.Module):
    def __init__(
        self, repoid_or_path, tokenizer: CLIPTokenizer, *args, **kwargs
    ) -> None:
        super().__init__()

        self.tokenizer = tokenizer

        self.vae: AutoencoderKL = AutoencoderKL.from_pretrained(
            repoid_or_path, *args, **kwargs, subfolder="vae"
        )
        self.unet: UNet2DConditionModel = UNet2DConditionModel.from_pretrained(
            repoid_or_path, *args, **kwargs, subfolder="unet"
        )
        self.text_encoder: CLIPTextModel = CLIPTextModel.from_pretrained(
            repoid_or_path, *args, **kwargs, subfolder="text_encoder"
        )

    def get_text_embeddings(self, text_tokens, n_imgs) -> torch.Tensor:
        text_ids = text_tokens.input_ids.to(self.text_encoder.device)

        text_embeddings = self.text_encoder(text_ids)[0]

        unconditional_tokens = self.text_tokenize([""] * len(text_ids))

        unconditional_ids = unconditional_tokens.input_ids.to(self.text_encoder.device)

        unconditional_embeddings = self.text_encoder(unconditional_ids)[0]

        text_embeddings = torch.repeat_interleave(
            torch.cat([unconditional_embeddings, text_embeddings]), n_imgs, dim=0
        )

        return text_embeddings

    def text_tokenize(self, prompts):
        return self.tokenizer(
            prompts,
            padding="max_length",
            max_length=self.tokenizer.model_max_length,
            truncation=True,
            return_tensors="pt",
        )

    def text_detokenize(self, tokens):
        return [
            self.tokenizer.decode(token)
            for token in tokens
            if token != self.tokenizer.vocab_size - 1
        ]

    def get_noise(self, batch_size, img_size) -> torch.Tensor:
        return torch.randn(
            (batch_size, self.unet.config.in_channels, img_size // 8, img_size // 8)
        )

    def get_initial_latents(self, n_imgs, img_size, n_prompts) -> torch.Tensor:
        latents = self.get_noise(n_imgs, img_size).repeat(n_prompts, 1, 1, 1)

        return latents

    def decode(self, latents):
        return self.vae.decode(1 / 0.18215 * latents).sample

    def encode(self, tensors):
        return self.vae.encode(tensors).latent_dist.mode() * 0.18215

    def predict_noise(
        self, scheduler, iteration, latents, text_embeddings, guidance_scale=7.5
    ):
        # expand the latents if we are doing classifier-free guidance to avoid doing two forward passes.
        latents = torch.cat([latents] * 2)
        latents = scheduler.scale_model_input(latents, scheduler.timesteps[iteration])

        # predict the noise residual
        noise_prediction = self.unet(
            latents,
            scheduler.timesteps[iteration],
            encoder_hidden_states=text_embeddings,
        ).sample

        # perform guidance
        noise_prediction_uncond, noise_prediction_text = noise_prediction.chunk(2)
        noise_prediction = noise_prediction_uncond + guidance_scale * (
            noise_prediction_text - noise_prediction_uncond
        )

        return noise_prediction

    def diffusion(
        self,
        scheduler,
        latents,
        text_embeddings,
        end_iteration=1000,
        start_iteration=0,
        **kwargs,
    ):
        for iteration in range(start_iteration, end_iteration):
            noise_pred = self.predict_noise(
                scheduler, iteration, latents, text_embeddings, **kwargs
            )

            # compute the previous noisy sample x_t -> x_t-1
            output = scheduler.step(noise_pred, scheduler.timesteps[iteration], latents)

            latents = output.prev_sample

        return latents


class DiffuserModel(AbstractModel):
    def __init__(self, *args, **kwargs) -> None:
        self.local_model: Diffuser = None
        self.meta_model: Diffuser = None
        self.tokenizer: CLIPTokenizer = None

        super().__init__(*args, **kwargs)

    def _register_increment_hook(self, hook) -> RemovableHandle:
        return self.local_model.unet.register_forward_hook(hook)

    def _load_meta(
        self, repoid_or_path, *args, device="cpu", **kwargs
    ) -> torch.nn.Module:
        self.tokenizer = CLIPTokenizer.from_pretrained(
            repoid_or_path, *args, **kwargs, subfolder="tokenizer"
        )

        return Diffuser(repoid_or_path, self.tokenizer, *args, **kwargs)

    def _load_local(
        self, repoid_or_path, *args, device="cpu", **kwargs
    ) -> torch.nn.Module:
        return Diffuser(repoid_or_path, self.tokenizer, *args, **kwargs).to(device)

    def _prepare_inputs(
        self,
        inputs,
        n_imgs=1,
        img_size=512,
    ) -> Any:
        if not isinstance(inputs, list):
            inputs = [inputs]

        latents = self.meta_model.get_initial_latents(n_imgs, img_size, len(inputs))

        text_tokens = self.meta_model.text_tokenize(inputs)

        return text_tokens, latents

    def _run_meta(self, inputs, *args, n_imgs=1, img_size=512, **kwargs) -> None:
        text_tokens, latents = self._prepare_inputs(
            inputs, n_imgs=n_imgs, img_size=img_size
        )

        text_embeddings = self.meta_model.get_text_embeddings(text_tokens, n_imgs)

        latents = torch.cat([latents] * 2).to("meta")

        self.meta_model.unet(
            latents,
            torch.zeros((1,), device="meta"),
            encoder_hidden_states=text_embeddings,
        ).sample

        self.meta_model.vae.decode(latents)

        return text_tokens.input_ids

    def _run_local(self, inputs, *args, n_imgs=1, img_size=512, **kwargs) -> None:
        text_tokens, latents = self._prepare_inputs(
            inputs, n_imgs=n_imgs, img_size=img_size
        )

        text_embeddings = self.meta_model.get_text_embeddings(text_tokens, n_imgs)

        latents = torch.cat([latents] * 2).to("meta")

        return self.meta_model.unet(
            latents,
            torch.zeros((1,), device="meta"),
            encoder_hidden_states=text_embeddings,
        ).sample

    def _generation(
        self,
        inputs,
        *args,
        n_steps=20,
        scheduler="LMSDiscreteScheduler",
        n_imgs=1,
        img_size=512,
        **kwargs,
    ) -> None:
        text_tokens, latents = self._prepare_inputs(
            inputs, n_imgs=n_imgs, img_size=img_size
        )

        text_embeddings = self.local_model.get_text_embeddings(text_tokens, n_imgs)

        if isinstance(scheduler, str):
            scheduler: SchedulerMixin = getattr(diffusers, scheduler).from_pretrained(
                self.repoid_or_path, subfolder="scheduler"
            )
        scheduler.set_timesteps(n_steps)

        latents = latents * scheduler.init_noise_sigma

        latents = self.local_model.diffusion(
            scheduler,
            latents.to(self.local_model.unet.device),
            text_embeddings.to(self.local_model.unet.device),
            *args,
            **kwargs,
            end_iteration=n_steps,
        )

        latents = (1 / 0.18215) * latents

        return self.local_model.vae.decode(latents).sample

    def to_image(self, latents) -> List[Image.Image]:
        """
        Function to convert latents to images
        """

        image = (latents / 2 + 0.5).clamp(0, 1)
        image = image.detach().cpu().permute(0, 2, 3, 1).numpy()
        images = (image * 255).round().astype("uint8")
        pil_images = [Image.fromarray(image) for image in images]

        return pil_images



import os

import yaml

from .pydantics.Config import ConfigModel
from .patching import *

PATH = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(PATH, "config.yaml"), "r") as file:
    CONFIG = ConfigModel(**yaml.safe_load(file))

from .models.DiffuserModel import DiffuserModel
from .models.LanguageModel import LanguageModel
from .models.AbstractModel import AbstractModel
from .Module import Module



import torch

from .. import util
from .Editor import Edit


class WrapperModuleEdit(Edit):
    def __init__(self, module_path: str, module_name: str) -> None:
        super().__init__()

        self.module_path = module_path
        self.module_name = module_name

        self.wrapper = util.WrapperModule()

    def edit(self, model: torch.nn.Module):
        module: torch.nn.Module = util.fetch_attr(model, self.module_path)
        setattr(module, self.module_name, self.wrapper)

    def restore(self, model: torch.nn.Module):
        module: torch.nn.Module = util.fetch_attr(model, self.module_path)
        delattr(module, self.module_name)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

import torch

from .. import util

if TYPE_CHECKING:
    from .Node import Node


class Proxy:
    """_summary_

    Attributes:
        node (Node): desc
    """

    def __init__(self, node: "Node") -> None:
        self.node = node

    def __call__(self, *args, **kwargs) -> Proxy:
        # If calling a method (not a sub-module) on the main module of this graph,
        # we want to trace into that method.
        if self.node.args[0] is self.node.graph.module_proxy.node and not isinstance(
            self.node.proxy_value, torch.nn.Module
        ):
            value = self.node.proxy_value.__func__(
                self.node.graph.module_proxy, *args, **kwargs
            )

            return value
        # Otherwise we just want to add a node saying we wish to call this module.
        else:
            value = self.node.proxy_value(
                *self.node.prepare_proxy_values(args),
                **self.node.prepare_proxy_values(kwargs),
            )

            return self.node.graph.add(
                graph=self.node.graph,
                value=value,
                target="__call__",
                args=[self.node] + list(args),
                kwargs=kwargs,
            )

    def __getitem__(self, key: Union[Proxy, Any]) -> Proxy:
        key = self.node.prepare_proxy_values(key)

        value = self.node.proxy_value[key]

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target="__getitem__",
            args=[self.node, key],
        )

    def __setitem__(self, key: Union[Proxy, Any], value: Union[Proxy, Any]) -> None:
        item_proxy = self[key]

        update = item_proxy.node.__class__.update

        update(item_proxy.node.proxy_value, item_proxy.node.prepare_proxy_values(value))

        item_proxy.node.graph.add(
            graph=item_proxy.node.graph,
            value=item_proxy.node.proxy_value,
            target=update,
            args=[item_proxy.node, value],
        )

    def __getattr__(self, key: Union[Proxy, Any]) -> Proxy:
        key = self.node.prepare_proxy_values(key)

        value = util.fetch_attr(self.node.proxy_value, key)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target=util.fetch_attr,
            args=[self.node, key],
        )

    def __len__(self) -> Proxy:
        value = len(self.node.proxy_value)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target=len,
            args=[self.node],
        )

    def __add__(self, other: Union[Proxy, Any]) -> Proxy:
        value = self.node.proxy_value + self.node.prepare_proxy_values(other)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target="__add__",
            args=[self.node, other],
        )

    def __sub__(self, other: Union[Proxy, Any]) -> Proxy:
        value = self.node.proxy_value - self.node.prepare_proxy_values(other)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target="__sub__",
            args=[self.node, other],
        )

    def __pow__(self, other: Union[Proxy, Any]) -> Proxy:
        value = self.node.proxy_value ** self.node.prepare_proxy_values(other)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target=pow,
            args=[self.node, other],
        )

    def __mul__(self, other: Union[Proxy, Any]) -> Proxy:
        value = self.node.proxy_value * self.node.prepare_proxy_values(other)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target="__mul__",
            args=[self.node, other],
        )

    def __truediv__(self, other: Union[Proxy, Any]) -> Proxy:
        value = self.node.proxy_value / self.node.prepare_proxy_values(other)

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target="__truediv__",
            args=[self.node, other],
        )

    def __bool__(self) -> bool:
        return self.node.proxy_value.__bool__()

    def __index__(self) -> int:
        return self.node.proxy_value.__index__()

    def __instancecheck__(self, __instance: Any) -> bool:
        return self.node.proxy_value.__instancecheck__(__instance)

    @classmethod
    def __torch_function__(cls, orig_method, types, args=None, kwargs=None) -> Proxy:
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()

        self: Proxy = args[0]

        value = orig_method(
            *self.node.prepare_proxy_values(args),
            **self.node.prepare_proxy_values(kwargs),
        )

        return self.node.graph.add(
            graph=self.node.graph,
            value=value,
            target=orig_method,
            args=args,
            kwargs=kwargs,
        )


from functools import wraps


def proxy_wrapper(fn) -> None:
    """Wraps problematic functions (torch functions sometimes).
    Checks if anty of its args are proxies. If so we return a proxy of the function.
    Otherwise just run the function.

    Args:
        fn (function): _description_

    Returns:
        _type_: _description_
    """

    @wraps(fn)
    def patched(*args, **kwargs):
        arguments = list(args) + list(kwargs.values())

        node = None

        for arg in arguments:
            if isinstance(arg, Proxy):
                node = arg.node

                break

        if node is not None:
            value = fn(
                *node.prepare_proxy_values(args),
                **node.prepare_proxy_values(kwargs),
            )

            return node.graph.add(
                graph=node.graph, value=value, target=fn, args=args, kwargs=kwargs
            )

        else:
            return fn(*args, **kwargs)

    return patched



from __future__ import annotations

from typing import Any, Callable, List, Tuple, Union

import torch.futures

from . import util
from .fx.Graph import Graph
from .fx.Proxy import Proxy


class TokenIndexer:
    def __init__(self, proxy: InterventionProxy) -> None:
        self.proxy = proxy

    def convert_idx(self, idx: int):
        if idx >= 0:
            n_tokens = self.proxy.node.proxy_value.shape[1]
            idx = -(n_tokens - idx)

        return idx

    def __getitem__(self, key: int) -> Proxy:
        key = self.convert_idx(key)

        return self.proxy[:, key]

    def __setitem__(self, key: int, value: Union[Proxy, Any]) -> None:
        key = self.convert_idx(key)

        self.proxy[:, key] = value


class InterventionProxy(Proxy):
    @staticmethod
    def proxy_save(value: Any) -> None:
        return util.apply(value, lambda x: x.clone(), torch.Tensor)

    def save(self) -> InterventionProxy:
        proxy = self.node.graph.add(
            graph=self.node.graph,
            value=self.node.proxy_value,
            target=InterventionProxy.proxy_save,
            args=[self.node],
        )

        self.node.graph.add(
            graph=self.node.graph,
            value=None,
            target="null",
            args=[proxy.node],
        )

        return proxy

    @property
    def token(self) -> TokenIndexer:
        return TokenIndexer(self)

    @property
    def t(self) -> TokenIndexer:
        return self.token

    @property
    def shape(self):
        return util.apply(self.node.proxy_value, lambda x: x.shape, torch.Tensor)

    @property
    def value(self):
        return self.node.future.value()


def intervene(activations, module_path: str, graph: Graph, key: str):
    """Entry to intervention graph. This should be hooked to all modules involved in intervention graph.

    Args:
        activations (_type_): _description_
        module_path (str): _description_
        graph (Graph): _description_
        key (str): _description_

    Returns:
        _type_: _description_
    """

    # Key to module activation argument nodes has format: <module path>.<output/input>.<generation index>.<batch index>
    module_path = f"{module_path}.{key}.{graph.generation_idx}"

    if module_path in graph.argument_node_names:
        argument_node_names = graph.argument_node_names[module_path]

        # multiple argument nodes can have same module_path if there are multiple invocations.
        for argument_node_name in argument_node_names:
            node = graph.nodes[argument_node_name]

            # args for argument nodes are (module_path, batch_size, batch_start)
            _, batch_size, batch_start = node.args

            # We set its result to the activations, indexed by only the relevant batch idxs.
            node.future.set_result(
                util.apply(
                    activations,
                    lambda x: x.narrow(0, batch_start, batch_size),
                    torch.Tensor,
                )
            )

    return activations


class HookModel:
    def __init__(
        self,
        model: torch.nn.Module,
        modules: List[str],
        input_hook: Callable = None,
        output_hook: Callable = None,
    ) -> None:
        self.model = model
        self.modules: List[Tuple[torch.nn.Module, str]] = [
            (util.fetch_attr(self.model, module_path), module_path)
            for module_path in modules
        ]
        self.input_hook = input_hook
        self.output_hook = output_hook

        self.handles = []

    def __enter__(self) -> HookModel:
        for module, module_path in self.modules:
            if self.input_hook is not None:

                def input_hook(module, input, module_path=module_path):
                    return self.input_hook(input, module_path)

                self.handles.append(module.register_forward_pre_hook(input_hook))

            if self.output_hook is not None:

                def output_hook(module, input, output, module_path=module_path):
                    return self.output_hook(output, module_path)

                self.handles.append(module.register_forward_hook(output_hook))

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        for handle in self.handles:
            handle.remove()





from abc import abstractmethod
from typing import Any


class Optimization:

    @abstractmethod
    def parameters(self):
        pass

    @abstractmethod
    def __call__(self) -> Any:
        pass


from __future__ import annotations
from abc import abstractmethod
from typing import List

import torch

class Edit:

    @abstractmethod
    def edit(self, obj: torch.nn.Module):
        pass

    @abstractmethod
    def restore(self, obj: torch.nn.Module):
        pass


class Editor:
    def __init__(self, obj: object, edits: List[Edit]) -> None:
        self.obj = obj
        self.edits = edits

    def __enter__(self) -> Editor:
        for edit in self.edits:
            edit.edit(self.obj)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        for edit in self.edits:
            edit.restore(self.obj)



from __future__ import annotations

import importlib
from typing import List


class Patch:
    def __init__(self, obj, replacement) -> None:
        self.obj = obj
        self.replacement = replacement

    def patch(self):
        module = importlib.import_module(self.obj.__module__)

        setattr(module, self.obj.__name__, self.replacement)

    def restore(self):
        module = importlib.import_module(self.obj.__module__)

        setattr(module, self.obj.__name__, self.obj)


class Patcher:
    def __init__(self, patches: List[Patch] = None) -> None:
        self.patches = patches or []

    def add(self, patch: Patch):
        self.patches.append(patch)

        patch.patch()

    def __enter__(self) -> Patcher:
        for patch in self.patches:
            patch.patch()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        for patch in self.patches:
            patch.restore()


DEFAULT_PATCHER = Patcher()

from functools import wraps

import torch


def repeat_interleave_wrapper(fn):
    @wraps(fn)
    def repeat_interleave(
        input: torch.Tensor, repeats: torch.LongTensor, dim=None, output_size=None
    ):
        if input.device.type == "meta":
            if not isinstance(repeats, torch.Tensor):
                repeats = torch.LongTensor([repeats])

            if dim is None:
                input = input.flatten()
                dim = 0

            if repeats.dim() == 0 or (repeats.dim() == 1 and repeats.size(0) == 1):
                repeats = repeats.reshape([1]).expand([input.size(dim)])

            new_dim_size = repeats.cumsum(0)[-1].item()
            new_output_shape = list(input.shape)
            new_output_shape[dim] = new_dim_size

            return torch.empty(new_output_shape, device="meta")

        else:
            return fn(input, repeats, dim=dim, output_size=output_size)

    return repeat_interleave


DEFAULT_PATCHER.add(
    Patch(torch.repeat_interleave, repeat_interleave_wrapper(torch.repeat_interleave))
)


DEFAULT_PATCHER.__enter__()



import nnsight
import pytest
import torch


@pytest.fixture(scope="module")
def gpt2(device: str):
    return nnsight.LanguageModel("gpt2", device_map=device)


@pytest.fixture
def MSG_prompt():
    return "Madison Square Garden is located in the city of"


def test_generation(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(max_new_tokens=3) as generator:
        with generator.invoke(MSG_prompt) as invoker:
            pass

    output = gpt2.tokenizer.decode(generator.output[0])

    assert output == "Madison Square Garden is located in the city of New York City"


def test_save(gpt2: nnsight.LanguageModel):
    with gpt2.generate(max_new_tokens=1) as generator:
        with generator.invoke("Hello world") as invoker:
            hs = gpt2.transformer.h[-1].output[0].save()

    assert hs.node.done()
    assert isinstance(hs.value, torch.Tensor)
    assert hs.value.ndim == 3


def test_set(gpt2: nnsight.LanguageModel):
    with gpt2.generate(max_new_tokens=1) as generator:
        with generator.invoke("Hello world") as invoker:
            pre = gpt2.transformer.h[-1].output[0].save()

            gpt2.transformer.h[-1].output[0] = 0

            post = gpt2.transformer.h[-1].output[0].save()

    output = gpt2.tokenizer.decode(generator.output[0])

    assert not (pre.value == 0).all().item()
    assert (post.value == 0).all().item()
    assert output != "Madison Square Garden is located in the city of New"


def test_adhoc_module(gpt2: nnsight.LanguageModel):
    with gpt2.generate() as generator:
        with generator.invoke("The Eiffel Tower is in the city of") as invoker:
            hidden_states = gpt2.transformer.h[-1].output[0]
            hidden_states = gpt2.lm_head(gpt2.transformer.ln_f(hidden_states))
            tokens = torch.softmax(hidden_states, dim=2).argmax(dim=2).save()

    output = gpt2.tokenizer.decode(tokens.value[0])

    assert output == "\n-el Tower is a the middle centre Paris"


def test_embeddings_set1(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(max_new_tokens=3) as generator:
        with generator.invoke(MSG_prompt) as invoker:
            embeddings = gpt2.transformer.wte.output

        with generator.invoke("_ _ _ _ _ _ _ _ _") as invoker:
            gpt2.transformer.wte.output = embeddings

    output1 = gpt2.tokenizer.decode(generator.output[0])
    output2 = gpt2.tokenizer.decode(generator.output[1])

    assert output1 == "Madison Square Garden is located in the city of New York City"
    assert output2 == "_ _ _ _ _ _ _ _ _ New York City"


def test_embeddings_set2(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(max_new_tokens=3) as generator:
        with generator.invoke(MSG_prompt) as invoker:
            embeddings = gpt2.transformer.wte.output.save()

    output1 = gpt2.tokenizer.decode(generator.output[0])

    with gpt2.generate(max_new_tokens=3) as generator:
        with generator.invoke("_ _ _ _ _ _ _ _ _") as invoker:
            gpt2.transformer.wte.output = embeddings.value

    output2 = gpt2.tokenizer.decode(generator.output[0])

    assert output1 == "Madison Square Garden is located in the city of New York City"
    assert output2 == "_ _ _ _ _ _ _ _ _ New York City"



from __future__ import annotations

from typing import Optional, Tuple, Union

import torch
from transformers.models import gpt2

from .. import util
from ..patching import Patch, Patcher


class GPT2AttentionAltered(gpt2.modeling_gpt2.GPT2Attention):
    def __init__(self, config, is_cross_attention=False, layer_idx=None):
        super().__init__(config, is_cross_attention, layer_idx)

        self.query = util.WrapperModule()
        self.key = util.WrapperModule()
        self.value = util.WrapperModule()

    def forward(
        self,
        hidden_states: Optional[Tuple[torch.FloatTensor]],
        layer_past: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = False,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]], ...]:
        if encoder_hidden_states is not None:
            if not hasattr(self, "q_attn"):
                raise ValueError(
                    "If class is used as cross attention, the weights `q_attn` have to be defined. "
                    "Please make sure to instantiate class with `GPT2Attention(..., is_cross_attention=True)`."
                )

            query = self.q_attn(hidden_states)
            key, value = self.c_attn(encoder_hidden_states).split(
                self.split_size, dim=2
            )
            attention_mask = encoder_attention_mask
        else:
            query, key, value = self.c_attn(hidden_states).split(self.split_size, dim=2)

        query = self._split_heads(query, self.num_heads, self.head_dim)
        key = self._split_heads(key, self.num_heads, self.head_dim)
        value = self._split_heads(value, self.num_heads, self.head_dim)

        # Altered -------------

        query = self.query(query)
        key = self.key(key)
        value = self.value(value)

        # ---------------------

        if layer_past is not None:
            past_key, past_value = layer_past
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if use_cache is True:
            present = (key, value)
        else:
            present = None

        if self.reorder_and_upcast_attn:
            attn_output, attn_weights = self._upcast_and_reordered_attn(
                query, key, value, attention_mask, head_mask
            )
        else:
            attn_output, attn_weights = self._attn(
                query, key, value, attention_mask, head_mask
            )

        attn_output = self._merge_heads(attn_output, self.num_heads, self.head_dim)
        attn_output = self.c_proj(attn_output)
        attn_output = self.resid_dropout(attn_output)

        outputs = (attn_output, present)
        if output_attentions:
            outputs += (attn_weights,)

        return outputs  # a, present, (attentions)


GPT2Patcher = Patcher([
    Patch(gpt2.modeling_gpt2.GPT2Attention, GPT2AttentionAltered)
    ])



from __future__ import annotations

from typing import Any, Type, Union

import torch

from . import util
from .contexts.Generator import Generator
from .fx.Graph import Graph
from .fx.Node import Node
from .intervention import InterventionProxy


class Module(torch.nn.Module):
    """_summary_

    Attributes:
        generator (Generator): _description_
        module_path (str): _description_
        output_shape (torch.Size): _description_
        output_type (Type): _description_
        _output (Proxy): _description_
    """

    def __init__(self) -> None:
        self.module_path: str = None
        self.input_shape: torch.Size = None
        self.input_type: Type = None
        self.output_shape: torch.Size = None
        self.output_type: Type = None

        self._output: InterventionProxy = None
        self._input: InterventionProxy = None
        self._graph: Graph = None

        self.generator: Generator = None

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """Override __call__ to check for InterventionProxy arguments. If there are any, we should return an
        InterventionProxy denoting we want to call the given module with arguments.

        Returns:
            Any: _description_
        """
        proxy = any(
            isinstance(x, InterventionProxy) for x in list(args) + list(kwds.values())
        )

        if proxy:
            module_proxy = getattr(self.generator.graph.module_proxy, self.module_path)

            return module_proxy.forward(*args, **kwds)

        return super().__call__(*args, **kwds)

    @property
    def output(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the output of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: _description_
        """
        if self._output is None:
            self._output = self.generator.graph.add(
                graph=self.generator.graph,
                value=util.apply(
                    self.output_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.output.{self.generator.generation_idx}",
                    self.generator.batch_size,
                    len(self.generator.prompts) - self.generator.batch_size,
                ],
            )

        return self._output

    @output.setter
    def output(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the output of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): _description_
        """

        Node.update(
            self.output.node.proxy_value, self.output.node.prepare_proxy_values(value)
        )

        self.output.node.graph.add(
            graph=self.output.node.graph,
            value=self.output.node.proxy_value,
            target=Node.update,
            args=[self.output.node, value],
        )
    
    @property
    def input(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the input of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: _description_
        """
        if self._input is None:
            self._input = self.generator.graph.add(
                graph=self.generator.graph,
                value=util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.input.{self.generator.generation_idx}",
                    self.generator.batch_size,
                    len(self.generator.prompts) - self.generator.batch_size,
                ],
            )

        return self._input

    @input.setter
    def input(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the input of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): _description_
        """

        Node.update(
            self.input.node.proxy_value, self.input.node.prepare_proxy_values(value)
        )

        self.input.node.graph.add(
            graph=self.input.node.graph,
            value=self.input.node.proxy_value,
            target=Node.update,
            args=[self.input.node, value],
        )

    @property
    def graph(self) -> Graph:
        if self._graph is None:
            self._graph = Graph.trace(
                self,
                *util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
            )

        return self._graph

    @staticmethod
    def wrap(module: torch.nn.Module) -> Module:
        """Wraps the torch Module with our Module

        Args:
            module (torch.nn.Module): _description_

        Returns:
            Module: _description_
        """

        def hook(module: Module, input: Any, output: Any):
            module._output = None
            module._input = None
            module.output_shape = util.apply(output, lambda x: x.shape, torch.Tensor)
            module.input_shape = util.apply(input, lambda x: x.shape, torch.Tensor)
            module.output_type = util.apply(output, lambda x: x.dtype, torch.Tensor)
            module.input_type = util.apply(input, lambda x: x.dtype, torch.Tensor)

        for name, _module in module.named_children():
            setattr(module, name, Module.wrap(_module))

        if isinstance(module, (Module, torch.nn.ModuleList)):
            return module

        util.wrap(module, Module)

        module.register_forward_hook(hook)

        return module



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

import torch.futures

from .. import util
from ..logger import logger
from .Proxy import Proxy

if TYPE_CHECKING:
    from .Graph import Graph


class Node:
    """_summary_

    Attributes:
        name (str): _description_
        graph (Graph): _description_
        proxy_value (Any): _description_
        target (Union[Callable, str]): _description_
        args (List[Any], optional): _description_. Defaults to None.
        kwargs (Dict[str, Any], optional): _description_. Defaults to None.
        meta (Dict[str, Any], optional): _description_. Defaults to None.
        listeners (List[Node]): desc
        dependencies (List[Node]): desc
        _future (torch.futures.Future): desc
        _proxy_device (torch.device): desc
    """

    @staticmethod
    def update(value1, value2) -> None:
        """Updates Tensor values with other Tensor values.

        Args:
            value1 (_type_): _description_
            value2 (_type_): _description_
        """
        if isinstance(value1, torch.Tensor):
            value1[:] = value2
        elif isinstance(value1, list) or isinstance(value1, tuple):
            for value_idx in range(len(value1)):
                Node.update(value1[value_idx], value2[value_idx])
        elif isinstance(value1, dict):
            for key in value1:
                Node.update(value1[key], value2[key])

    @staticmethod
    def target_name(target) -> str:
        if isinstance(target, str):
            name = target
        elif callable(target):
            name = target.__name__

        return name

    def __init__(
        self,
        name: str,
        graph: "Graph",
        value: Any,
        target: Union[Callable, str],
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
        meta: Dict[str, Any] = None,
    ) -> None:
        super().__init__()

        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if meta is None:
            meta = dict()

        self.name = name
        self.graph = graph
        self.proxy_value = value
        self.target = target
        self.args = util.apply(args, lambda x: x.node, Proxy)
        self.kwargs = util.apply(kwargs, lambda x: x.node, Proxy)
        self.meta = meta

        self.listeners: List[Node] = list([self])
        self.dependencies: List[Node] = list()

        # Add all arguments that are nodes to nodes dependencies
        util.apply(self.args, lambda x: self.dependencies.append(x), Node)
        util.apply(self.kwargs, lambda x: self.dependencies.append(x), Node)
        # Add node to all arguments that are nodes' listeners
        util.apply(self.args, lambda x: x.listeners.append(self), Node)
        util.apply(self.kwargs, lambda x: x.listeners.append(self), Node)

        self._future: torch.futures.Future = None
        self._proxy_device: torch.device = None

    @property
    def future(self) -> torch.futures.Future:
        """Lazy creation of _future attribute.

        Returns:
            torch.futures.Future: _description_
        """
        if self._future is None:
            self._future = torch.futures.Future()

        return self._future

    @property
    def proxy_device(self) -> torch.device:
        """Lazy creation of _proxy_device attribute.

        Returns:
            torch.Device: _description_
        """
        if self._proxy_device is None:
            device = None

            def _device(value):
                nonlocal device
                device = value.device

            util.apply(self.proxy_value, _device, torch.Tensor)
            # TODO
            # util.apply(self.proxy_value, _device, torch.nn.Module)

            self._proxy_device = device

        return self._proxy_device

    def prepare_proxy_values(self, values):
        def slice_to_value(arg: slice):
            return slice(
                self.prepare_proxy_values(arg.start),
                self.prepare_proxy_values(arg.stop),
                self.prepare_proxy_values(arg.step),
            )

        # Convert procies to their proxy_value
        values = util.apply(values, lambda x: x.node.proxy_value, Proxy)
        # Slices may have proxies as part of their attributes so convert those to their proxy_values
        values = util.apply(values, slice_to_value, slice)
        # Move tensors to that of the proxy_device (probably 'meta')
        values = util.apply(values, lambda x: x.to(self.proxy_device), torch.Tensor)

        return values

    def compile(self) -> None:
        # When this future is done, log that event.
        self.future.add_done_callback(lambda x: logger.debug(f"=> SET({self.name})"))

        # Nodes tell listeners when to try and be executed.
        # This chains futures so after this node's future is done, it goes through
        # it's listeners in order and calls their .chain() method.
        future = self.listeners[0].future

        for listener in self.listeners[1:]:
            future = future.then(listener.chain)

        # Collect all listeners futures into a single future that when done, call this
        # nodes .destroy() method.
        torch.futures.collect_all(
            util.apply(self.listeners, lambda x: x.future, Node)
        ).add_done_callback(lambda x: self.destroy())

    def value(self) -> Any:
        """Wrapper for this node's future .value()

        Returns:
            Any: _description_
        """
        return self.future.value()

    def done(self) -> bool:
        """Wrapper for this node's future .done()

        Returns:
            bool: _description_
        """
        return self.future.done()

    def fufilled(self) -> bool:
        """Returns True if all of this node's dependencies are done.

        Returns:
            bool: _description_
        """
        for dependency in self.dependencies:
            if not dependency.done():
                return False

        return True

    def prepare_inputs(self) -> Tuple[List[Any], Dict[str, Any]]:
        # Turn futures into their value
        def _value(value: Node):
            return value.value()

        args = util.apply(self.args, _value, Node)
        kwargs = util.apply(self.kwargs, _value, Node)

        device = None

        def _device(value):
            nonlocal device
            device = value.device

        all_args = list(args) + list(kwargs.values())

        util.apply(list(reversed(all_args)), _device, torch.Tensor)
        # util.apply(list(reversed(all_args)), _device, torch.nn.Module)

        # Move tensors to device
        def _to(value: torch.Tensor):
            return value.to(device)

        args = util.apply(args, _to, torch.Tensor)
        kwargs = util.apply(kwargs, _to, torch.Tensor)

        return args, kwargs

    def execute(self) -> None:
        """Actually executes this node."""

        # We se a nodes target to 'null' if we don't want it to be executed and therefore never done
        if self.target == "null":
            return

        # Prepare arguments.
        args, kwargs = self.prepare_inputs()

        # If target is a string, it must be a method attribute on the first argument object.
        if isinstance(self.target, str):
            obj, *args = args

            target = getattr(obj, self.target)
        # Otherwise it must be the function itself.
        else:
            target = self.target

        # Call the target to get value.
        output = target(*args, **kwargs)

        # Set this nodes future value to result.
        self.future.set_result(output)

    def destroy(self) -> None:
        """Removes the reference to the node's _future and logs it's destruction."""
        logger.debug(f"=> DEL({self.name})")

        self._future = None

    def chain(self, future: torch.futures.Future):
        # If all of a nodes dependencies are done, execute it.
        # Dont execute if already done.
        if self.fufilled() and not self.done():
            try:
                self.execute()
            except Exception as e:
                # TODO
                # An exectption is actually never thrown upward to the point it stops the program. Need to find a way.
                logger.exception(f"Exception in execution of node '{self.name}'.")

                self.future.set_exception(e)
                future.set_exception(e)

                raise e

        future.set_result(None)

    def __str__(self) -> str:
        args = util.apply(self.args, lambda x: f"'{x}'", str)
        args = util.apply(args, lambda x: x.name, Node)
        args = [str(arg) for arg in args]
        meta = f"{self.meta['file']}({self.meta['line']})" if self.meta else ""
        return f"{self.name}:[ {meta} args:({','.join(args)}) l:{len(self.listeners)} d:{len(self.dependencies)}]"



from __future__ import annotations

import inspect
from typing import Any, Callable, Dict, List, Type, Union

import torch

from .. import util
from .Node import Node
from ..patching import Patcher, Patch
from .Proxy import Proxy, proxy_wrapper


class Graph:
    """Represents a computation graph involving a Module

    Attributes:
        proxy_class (Type[Proxy]): Proxy class to use. Defaults to Proxy.
        nodes (Dict[str, Node]): Mapping of node name to node.
        name_idx (Dict[str, int]): Mapping of node target_name to number of previous names with the same target_name.
            Used so names are unique.
        module_proxy (Proxy): Proxy for given root module
        argument_node_names (Dict[str, List[str]]): _description_
        generation_idx (int): _description_

    """

    @staticmethod
    def trace(
        module: torch.nn.Module, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> Graph:
        """Given a module and some default (should be meta tensors) arguments, create a graph from the module's
        forward method.

        Args:
            module (torch.nn.Module): _description_
            args (List[Any]): desc
            kwargs (Dict[str, Any]): desc

        Returns:
            Graph: _description_
        """

        # Create a graph with the module as the root module
        graph = Graph(module)

        # Get 'unbound' version of forward method so we can pass in proxy of module insead of self
        forward = module.__class__.forward

        # Want list not tuple
        args = list(args)

        # Inspect forward signature to collect all parameters
        signature = inspect.signature(forward)

        def get_argument_value(param: inspect.Parameter, idx: int):
            """Gets the correct argument to pass to forward method.


            Args:
                param (_type_): _description_
                idx (_type_): _description_

            Returns:
                _type_: _description_
            """

            # If idx in range of provided args, create a proxy for that arg instead of default.
            if idx < len(args):
                return graph.add(
                    graph=graph, value=args[idx], target="argument", args=[param.name]
                )
            # If param name in provided kwargs, create a proxy for that arg instead of default.
            if param.name in kwargs:
                return graph.add(
                    graph=graph,
                    value=kwargs[param.name],
                    target="argument",
                    args=[param.name],
                )
            # Otherwise just return default
            return param.default

        # Create the appropriate proxies/values for the forward method in order to trace.
        arguments = [
            get_argument_value(param, i)
            for i, param in enumerate(list(signature.parameters.values())[1:])
        ]

        # Some methods cannot be caught because they arent torch functions or dont play nice with __torch_function__.
        # So the patcher repalces the methods with something to catch proxies and return proxies.
        with Patcher() as patcher:
            patcher.add(Patch(torch.full, proxy_wrapper(torch.full)))
            patcher.add(Patch(torch.finfo, proxy_wrapper(torch.finfo)))
            patcher.add(Patch(torch.arange, proxy_wrapper(torch.arange)))

            # Run forward with root module proxy and arguments
            output: Proxy = forward(graph.module_proxy, *arguments)

            # Get proxy_value for return
            value = util.apply(output, lambda x: x.node.proxy_value, Proxy)

            # Create the 'rtn_0' return proxy
            return_proxy = graph.add(
                graph=graph, value=value, target=Graph.rtn, args=output
            )

            # This is how we tell the graph not to destroy a proxy after it's listeners are completed.
            # Create a 'null' proxy. The return proxy listens to the 'null' proxy with args=[return_proxy.node] but 'null' will never be completed.
            graph.add(
                graph=graph,
                value=None,
                target="null",
                args=[return_proxy.node],
            )

        return graph

    @staticmethod
    def rtn(*args, **kwargs):
        """
        Function to just pass through data for returning data in a graph forward method.

        Returns:
            _type_: _description_
        """

        return args

    def __init__(
        self, module: torch.nn.Module, proxy_class: Type[Proxy] = Proxy
    ) -> None:
        """_summary_

        Args:
            module (torch.nn.Module): _description_
            proxy_class (Type[Proxy], optional): _description_.
        """
        self.proxy_class = proxy_class

        self.nodes: Dict[str, Node] = dict()
        self.name_idx: Dict[str, int] = dict()

        self.module_proxy = self.add(graph=self, value=module, target="module")
        self.argument_node_names: Dict[str, List[str]] = dict()

        self.generation_idx = 0

    def increment(self) -> None:
        """Increments the generation_idx by one. Should be called by a forward hook on the model being used for generation."""
        self.generation_idx += 1

    def compile(self, module: torch.nn.Module) -> None:
        """Re-compile graph to prepare for a new execution of the graph.

        Args:
            module (torch.nn.Module): Module to be considered the root module of the graph.
        """

        # Remove nodes that have no effect.
        self.eliminate_dead_code()

        # Reset all node futures.
        for node in self.nodes.values():
            node._future = None
        # Compile nodes individually.
        for node in self.nodes.values():
            node.compile()

        self.generation_idx = 0

        # Setting the root module future kicks off the graph execution.
        self.nodes["module_0"].future.set_result(module)

    def add(
        self,
        graph: Graph,
        value: Any,
        target: Union[Callable, str],
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
        name: str = None,
    ) -> Proxy:
        """Adds a node to the graph and returns it's proxy.

        Args:
            graph (Graph): _description_
            value (Any): 'meta' proxy value used for tracing the shapes and values.
            target (Union[Callable, str]): Either the function to call for this node, or a string that's the name of a method attribute on the first arg.
            args (List[Any], optional): _description_. Defaults to None.
            kwargs (Dict[str, Any], optional): _description_. Defaults to None.
            name (str, optional): _description_. Defaults to None.

        Returns:
            Proxy: _description_
        """
        target_name = Node.target_name(target)

        if target_name not in self.name_idx:
            self.name_idx[target_name] = 0
        else:
            if target_name == "rtn":
                raise ValueError("Can only have one return ('rtn') node.")
            if target_name == "module":
                raise ValueError("Can only have one module node.")

        if name is None:
            name = f"{target_name}_{self.name_idx[target_name]}"

        self.name_idx[target_name] += 1

        stack = inspect.stack()
        proxy_frame = stack[2]

        node = Node(
            name=name,
            graph=graph,
            value=value,
            target=target,
            args=args,
            kwargs=kwargs,
            meta={"line": proxy_frame.lineno, "file": proxy_frame.filename},
        )

        self.nodes[name] = node

        if target_name == "argument":
            module_path = args[0]

            if module_path not in self.argument_node_names:
                self.argument_node_names[module_path] = []

            self.argument_node_names[module_path].append(name)

        return self.proxy(node)

    def proxy(self, node: Node) -> Proxy:
        """Returns proxy of node with specified proxy_class.

        Args:
            node (Node): _description_

        Returns:
            Proxy: _description_
        """
        return self.proxy_class(node)

    def eliminate_dead_code(self):
        # TODO
        pass

    def wrap(self, module: torch.nn.Module) -> torch.nn.Module:
        """Replaces the forward method of the given module with an execution of the module's graph.

        Args:
            module (torch.nn.Module): _description_

        Returns:
            torch.nn.Module: _description_
        """

        def forward(*args, **kwargs):
            # Compile the graph with the given module as the root module.
            self.compile(module)

            # Gets list of all argument nodes for this graph.
            argument_nodes_list = list(self.argument_node_names.values())

            # Sets the result of the argument nodes future for args.
            for i, arg in enumerate(args):
                self.nodes[argument_nodes_list[i][0]].future.set_result(arg)

            # And then for kwargs.
            for key in kwargs:
                if key in self.argument_node_names:
                    self.nodes[self.argument_node_names[key][0]].future.set_result(arg)

            # 'rtn_0' should have the value we need to return.
            return_value = self.nodes["rtn_0"].value()
            self.nodes["rtn_0"].destroy()
            return return_value

        # Repalce forward method with custom graph execution method.
        module.forward = forward

        return module

    def __str__(self) -> str:
        result = ""

        for name, node in self.nodes.items():
            result += f"  %{node}\n"

        return result



import logging
import os

PATH = os.path.dirname(os.path.abspath(__file__))
logging_handler = logging.FileHandler(os.path.join(PATH, f"nnsight.log"), "a")
logging_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s %(processName)-10s %(name)s %(levelname)-8s %(message)s"
    )
)
logging_handler.setLevel(logging.DEBUG)
logger = logging.getLogger("nnsight")
logger.addHandler(logging_handler)
logger.setLevel(logging.DEBUG)



from __future__ import annotations

import pickle
from typing import TYPE_CHECKING, Dict, List, Union

import socketio

from .. import CONFIG, pydantics
from ..fx.Graph import Graph
from ..intervention import InterventionProxy
from .Invoker import Invoker

if TYPE_CHECKING:
    from ..models.AbstractModel import AbstractModel


class Generator:
    """_summary_

    Attributes:
        model (Model): Model object this is a generator for.
        blocking (bool): If when using device_map='server', block and wait form responses. Otherwise have to manually
            request a response.
        args (List[Any]): Arguments for calling the model.
        kwargs (Dict[str,Any]): Keyword arguments for calling the model.
        generation_idx (int): Keeps track of what iteration of generation to do interventions at. Used by the Module class
            to specify generation_idx for interventions and changed by the Invoker class using invoker.next().
        batch_size (int): Current size of invocation batch. To be used by Module node creation
        prompts (List[str]): Keeps track of prompts used by invokers.
        graph (Graph): Graph of all user intervention operations.
        output (??): desc
    """

    def __init__(
        self,
        model: "AbstractModel",
        *args,
        blocking: bool = True,
        server: bool = False,
        **kwargs,
    ) -> None:
        self.model = model
        self.server = server
        self.blocking = blocking
        self.args = args
        self.kwargs = kwargs

        self.graph = Graph(self.model.meta_model, proxy_class=InterventionProxy)

        self.generation_idx: int = 0
        self.batch_size: int = 0
        self.prompts: List[str] = []
        self.output = None

        # Modules need to know about the current generator to create the correct proxies.
        for name, module in self.model.named_modules():
            module.generator = self

    def __enter__(self) -> Generator:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """On exit, run and generate using the model whether locally or on the server."""
        if self.server:
            self.run_server()
        else:
            self.run_local()

    def run_local(self):
        # Run the model and store the output.
        self.output = self.model(self.model._generation, self.prompts, self.graph, *self.args, **self.kwargs)

    def run_server(self):
        # Create the pydantic class for the request.
        request = pydantics.RequestModel(
            args=self.args,
            kwargs=self.kwargs,
            model_name=self.model.model_name_or_path,
            prompts=self.prompts,
            intervention_graph=self.graph,
        )

        if self.blocking:
            self.blocking_request(request)
        else:
            self.non_blocking_request(request)

    def blocking_request(self, request: pydantics.RequestModel):
        # Create a socketio connection to the server.
        sio = socketio.Client()
        sio.connect(f"ws://{CONFIG.API.HOST}")

        # Called when recieving a response from the server.
        @sio.on("blocking_response")
        def blocking_response(data):
            # Load the data into the ResponseModel pydantic class.
            data: pydantics.ResponseModel = pickle.loads(data)

            # Print response for user ( should be logger.info and have an infor handler print to stdout)
            print(str(data))

            # If the status of the response is completed, update the local futures that the user specified to save.
            # Then disconnect and continue.
            if data.status == pydantics.JobStatus.COMPLETED:
                for name, value in data.saves.items():
                    self.graph.nodes[name].future.set_result(value)

                self.output = data.output

                sio.disconnect()
            # Or if there was some error.
            elif data.status == pydantics.JobStatus.ERROR:
                sio.disconnect()

        sio.emit(
            "blocking_request",
            request.model_dump(exclude_defaults=True, exclude_none=True),
        )

        sio.wait()

    def non_blocking_request(self, request: pydantics.RequestModel):
        pass

    def invoke(self, input, *args, **kwargs) -> Invoker:
        return Invoker(self, input, *args, **kwargs)



from __future__ import annotations

import pickle
from datetime import datetime
from typing import Dict, List, Type, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer
)

from ..fx.Graph import Graph
from .fx import NodeModel


class RequestModel(BaseModel):
    model_config = ConfigDict(protected_namespaces=(), arbitrary_types_allowed=True)

    args: List
    kwargs: Dict
    model_name: str
    prompts: List[str]
    intervention_graph: Union[Graph, bytes, Dict[str, NodeModel]]
    # Edits
    # altered

    id: str = None
    recieved: datetime = None
    blocking: bool = False

    @field_serializer("intervention_graph")
    def intervention_graph_serialize(self, value: Union[str, Graph], _info) -> str:
        if isinstance(value, Graph):
            nodes = dict()

            for node in value.nodes.values():
                node = NodeModel.from_node(node)
                nodes[node.name] = node

            value = nodes

        return pickle.dumps(value)

    def graph(self):

        graph = Graph(None)

        for node in self.intervention_graph.values():
            NodeModel.to_node(graph, self.intervention_graph, node)

        return graph



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable, List, Union

import accelerate
import torch
from torch.utils.hooks import RemovableHandle

from ..alteration import REPOID_TO_ALTERATION
from ..contexts.Generator import Generator
from ..contexts.Runner import Runner
from ..editing.Editor import Edit, Editor
from ..editing.GraphEdit import GraphEdit
from ..editing.WrapperModuleEdit import WrapperModuleEdit
from ..fx.Graph import Graph
from ..intervention import intervene, HookModel
from ..logger import logger
from ..Module import Module
from ..patching import Patcher


class AbstractModel:
    """_summary_
    """
    def __init__(self, repoid_or_path:str, *args, alter:bool=True, **kwargs) -> None:
        super().__init__()

        # TODO handle passing in a torch module
        self.repoid_or_path = repoid_or_path
        self.args = args
        self.kwargs = kwargs
        # Boolean on whether to check if alterations exist for this module and apply them.
        self.alter = alter
        # Boolean on whether this model has been dispatched (locally loaded) yet
        self.dispatched = False
        self.local_model: torch.nn.Module = None
        self.edits: List[Edit] = list()

        logger.debug(f"Initializing `{self.repoid_or_path}`...")

        # If alter and alteration exist, use alteration patcher context while loading module.
        with self.alteration() if self.alter else Patcher():
            # Use accelerate and .to('meta') to assure tensors are loaded to 'meta' device
            with accelerate.init_empty_weights(include_buffers=True):
                self.meta_model: torch.nn.Module = Module.wrap(
                    self._load_meta(self.repoid_or_path, *args, **kwargs).to("meta")
                )

        # Wrap all modules in our Module class.
        for name, module in self.meta_model.named_children():
            module = Module.wrap(module)

            setattr(self.meta_model, name, module)

        # Set module_path attribute so Modules know their place.
        for name, module in self.meta_model.named_modules():
            module.module_path = name

        # Run inital dummy string to populate Module shapes, dtypes etc
        self._run_meta("_")

        logger.debug(f"Initialized `{self.repoid_or_path}`")

    def __repr__(self) -> str:
        return repr(self.meta_model)

    def __getattr__(self, key) -> Any:
        """Allows access of sub-modules on meta_model directly from AbstractModel object

        Args:
            key (_type_): _description_

        Returns:
            Any: _description_
        """
        return getattr(self.meta_model, key)

    def __call__(
        self,
        fn: Callable,
        inputs: Any,
        graph: Graph,
        *args,
        edits: List[Edit] = None,
        inference: bool = True,
        **kwargs,
    ) -> Any:
        """Runs some function with some inputs and some graph with the approriate context for this model.

        Args:
            fn (Callable): _description_
            inputs (Any): _description_
            graph (Graph): _description_
            edits (List[Edit], optional): _description_. Defaults to None.
            inference (bool, optional): _description_. Defaults to True.

        Returns:
            Any: _description_
        """
        if edits is None:
            edits = self.edits


        # If local_model not yet loaded, do so.
        if not self.dispatched:
            with self.alteration() if self.alter else Patcher():
                self.local_model = self._load_local(
                    self.repoid_or_path, *self.args, **self.kwargs
                )

                # By default, all params should be frozen.
                for param in self.local_model.parameters():
                    param.requires_grad = False


        with Editor(self, edits):

            # Send local_model to graph to re-compile
            graph.compile(self.local_model)

            increment_hook = self._register_increment_hook(
                lambda module, input, output: graph.increment()
            )

            # The intervention graph for running a Model will have the modules that are involved
            # in the graph's argument_node_names.
            modules = set(
                [
                    ".".join(name.split(".")[:-2])
                    for name in graph.argument_node_names.keys()
                ]
            )

            logger.debug(f"Running `{self.repoid_or_path}`...")

            self.local_model.eval() if inference else self.local_model.train()

            with torch.inference_mode(mode=inference):
                with HookModel(
                    self.local_model,
                    list(modules),
                    input_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "input"
                    ),
                    output_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "output"
                    ),
                ):
                    output = fn(inputs, *args, **kwargs)

            increment_hook.remove()

            logger.debug(f"Completed `{self.repoid_or_path}`")

        return output

    def alteration(self) -> Patcher:
        return REPOID_TO_ALTERATION.get(self.repoid_or_path, Patcher())

    def generate(self, *args, **kwargs) -> Generator:
        return Generator(self, *args, **kwargs)

    def forward(self, inputs, *args, **kwargs) -> Runner:
        return Runner(self, inputs, *args, **kwargs)

    def modulize(self, module: Module, node_name: str, module_name: str) -> None:
        """_summary_

        Args:
            module (Module): _description_
            node_name (str): _description_
            module_name (str): _description_
        """

        # Create a WrapperModuleEdit which just adds a WrapperModule to an existing module at the given moduel_name.
        wme = WrapperModuleEdit(module.module_path, module_name)
        # Wrap with our Module and update new attributes.
        wme.wrapper: Module = Module.wrap(wme.wrapper)
        wme.wrapper.module_path = f"{module.module_path}.{module_name}"
        wme.wrapper.generator = module.generator
        wme.wrapper.output_shape = module.output_shape
        # Carry out the edit on the meta_model.
        wme.edit(self.meta_model)

        # Get/create the execution graph for the module's forward method.
        graph = module.graph

        # Add two proxies/nodes, one to get the new WrapperModule we added and another to call it with the data from the original module.
        # Passing the data through the wrapper module allows hooking of the module's output like usual.
        module_proxy = getattr(graph.module_proxy, module_name)
        module_proxy(graph.nodes[node_name])

        # Create and carry out the edit on the meta_model.
        ge = GraphEdit(module.module_path, module.graph)
        ge.edit(self.meta_model)

        # Append to self.edits so when we call the local model, we temporarily edit the module in the same way as the meta model.
        self.edits.append(wme)
        self.edits.append(ge)

    @abstractmethod
    def _prepare_inputs(self, inputs: Any, **kwargs) -> Any:
        """Abstract method for Model type to process inputs.

        Args:
            inputs (Any): _description_

        Returns:
            Any: _description_
        """
        raise NotImplementedError()

    @abstractmethod
    def _load_meta(self, repoid_or_path, *args, **kwargs) -> torch.nn.Module:
        """Abstract method for Model type to initialize what it needs for it's meta model.

        Args:
            repoid_or_path (_type_): _description_

        Returns:
            torch.nn.Module: _description_
        """
        raise NotImplementedError()

    @abstractmethod
    def _load_local(self, repoid_or_path, *args, **kwargs) -> torch.nn.Module:
        raise NotImplementedError()

    @abstractmethod
    def _run_meta(self, inputs, *args, **kwargs) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def _run_local(self, inputs, *args, **kwargs) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def _generation(self, inputs, *args, **kwargs) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def _register_increment_hook(self, hook) -> RemovableHandle:
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union

from ..fx.Graph import Graph
from ..intervention import InterventionProxy

if TYPE_CHECKING:
    from ..models.AbstractModel import AbstractModel


# TODO make parent class for Runner and Generator as Module depends on attributes
class Runner:
    def __init__(
        self,
        model: "AbstractModel",
        input,
        *args,
        inference=False,
        **kwargs,
    ) -> None:
        self.model = model

        self.input = input
        self.inference = inference
        self.args = args
        self.kwargs = kwargs

        self.graph = Graph(self.model.meta_model, proxy_class=InterventionProxy)

        self.batch_size: int = 0
        self.prompts: List[str] = []
        self.generation_idx = 0
        self.output = None

        # Modules need to know about the current generator to create the correct proxies.
        for name, module in self.model.named_modules():
            module.generator = self

    def __enter__(self) -> Runner:
        token_ids = self.model._run_meta(self.input, *self.args, **self.kwargs)

        # Decode tokenized inputs for user usage.
        self.tokens = [
            [self.model.tokenizer.decode(token) for token in ids] for ids in token_ids
        ]
        self.ids = token_ids

        self.batch_size = len(self.ids)

        self.prompts.extend(["".join(tokens) for tokens in self.tokens])

        if len(self.tokens) == 1:
            self.tokens = self.tokens[0]
            self.ids = self.ids[0]

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.output = self.model(
            self.model._run_local,
            self.input,
            self.graph,
            *self.args,
            inference=self.inference,
            **self.kwargs,
        )



from typing import Any
import torch

from ...Module import Module
from . import Optimization


class LORA(Optimization):
    def __init__(self, module: Module, r: int) -> None:
        self.module = module
        self.r = r

        self.WA = torch.nn.Parameter(torch.empty(self.module.input_shape[0][-1], self.r), requires_grad=True)
        self.WB = torch.nn.Parameter(torch.empty(self.r, self.module.output_shape[-1]), requires_grad=True)

    def __call__(self, alpha:float=1.0) -> Any:

        inp = self.module.input[0]

        self.module.output = (torch.matmul(torch.matmul(inp, self.WA), self.WB) + self.module.output) * alpha

    def parameters(self):
        return [self.WA, self.WB]



from __future__ import annotations

import logging
from datetime import datetime
from enum import Enum
from typing import Any, Dict

from pydantic import BaseModel
from transformers.generation.utils import GenerateOutput


class JobStatus(Enum):
    RECIEVED = "RECIEVED"
    APPROVED = "APPROVED"
    SUBMITTED = "SUBMITTED"
    COMPLETED = "COMPLETED"

    ERROR = "ERROR"


class ResponseModel(BaseModel):
    id: str
    status: JobStatus
    description: str

    output: Any = None
    recieved: datetime = None
    saves: Dict[str, Any] = None
    blocking: bool = False

    def __str__(self) -> str:
        return f"{self.id} - {self.status.name}: {self.description}"

    def log(self, logger: logging.Logger) -> ResponseModel:
        if self.status == JobStatus.ERROR:
            logger.error(str(self))
        else:
            logger.info(str(self))

        return self



from .Editor import Edit
import torch
from .. import util
from ..fx.Graph import Graph
class GraphEdit(Edit):

    def __init__(self, module_path:str, graph:Graph) -> None:
        super().__init__()

        self.module_path = module_path
        self.graph = graph

        self.forward = None

    def edit(self, model:torch.nn.Module):
        module: torch.nn.Module = util.fetch_attr(model, self.module_path)
        self.forward = module.forward
        self.graph.wrap(module)

    def restore(self, model:torch.nn.Module):
        module: torch.nn.Module = util.fetch_attr(model, self.module_path)
        setattr(module, 'forward', self.forward)


from typing import Any
import torch

from ...Module import Module
from . import Optimization


class SoftPrompt(Optimization):
    def __init__(self, module: Module, n: int) -> None:
        self.module = module
        self.n = n

        self.embedding = torch.nn.Parameter(
            torch.zeros((self.n, self.module.embedding_dim)), requires_grad=True
        )

    def __call__(self) -> Any:
        self.module.output = self.embedding[:]

    def parameters(self):
        return [self.embedding]
